@@ -61,6 +61,16 @@ static int set_peer(struct wireguard_device *wg, void __user *user_peer, size_t
 	if (!peer) { /* Peer doesn't exist yet. Add a new one. */
 		if (in_peer.flags & WGPEER_REMOVE_ME)
 			return -ENODEV; /* Tried to remove a non existing peer. */
+
+		down_read(&wg->static_identity.lock);
+		if (wg->static_identity.has_identity && !memcmp(in_peer.public_key, wg->static_identity.static_public, NOISE_PUBLIC_KEY_LEN)) {
+			/* We silently ignore peers that have the same public key as the device. The reason we do it silently
+			 * is that we'd like for people to be able to reuse the same set of API calls across peers. */
+			up_read(&wg->static_identity.lock);
+			goto out;
+		}
+		up_read(&wg->static_identity.lock);
+
 		peer = peer_rcu_get(peer_create(wg, in_peer.public_key));
 		if (!peer)
 			return -ENOMEM;
@@ -75,7 +85,7 @@ static int set_peer(struct wireguard_device *wg, void __user *user_peer, size_t
 	}
 
 	if (in_peer.endpoint.addr.sa_family == AF_INET || in_peer.endpoint.addr.sa_family == AF_INET6) {
-		struct endpoint endpoint = { 0 };
+		struct endpoint endpoint = { { { 0 } } };
 		memcpy(&endpoint, &in_peer.endpoint, sizeof(in_peer.endpoint));
 		socket_set_peer_endpoint(peer, &endpoint);
 	}
@@ -109,7 +119,7 @@ out:
 
 int config_set_device(struct wireguard_device *wg, void __user *user_device)
 {
-	int ret = 0;
+	int ret;
 	size_t i, offset;
 	struct wgdevice in_device;
 	void __user *user_peer;
@@ -120,10 +130,13 @@ int config_set_device(struct wireguard_device *wg, void __user *user_device)
 
 	mutex_lock(&wg->device_update_lock);
 
-	if (copy_from_user(&in_device, user_device, sizeof(in_device))) {
-		ret = -EFAULT;
+	ret = -EFAULT;
+	if (copy_from_user(&in_device, user_device, sizeof(in_device)))
+		goto out;
+
+	ret = -EPROTO;
+	if (in_device.version_magic != WG_API_VERSION_MAGIC)
 		goto out;
-	}
 
 	if (in_device.fwmark || (!in_device.fwmark && (in_device.flags & WGDEVICE_REMOVE_FWMARK))) {
 		wg->fwmark = in_device.fwmark;
@@ -143,6 +156,16 @@ int config_set_device(struct wireguard_device *wg, void __user *user_device)
 		noise_set_static_identity_private_key(&wg->static_identity, NULL);
 		modified_static_identity = true;
 	} else if (memcmp(zeros, in_device.private_key, WG_KEY_LEN)) {
+		u8 public_key[NOISE_PUBLIC_KEY_LEN] = { 0 };
+		struct wireguard_peer *peer;
+		/* We remove before setting, to prevent race, which means doing two 25519-genpub ops. */
+		bool unused __attribute((unused)) = curve25519_generate_public(public_key, in_device.private_key);
+		peer = pubkey_hashtable_lookup(&wg->peer_hashtable, public_key);
+		if (peer) {
+			peer_put(peer);
+			peer_remove(peer);
+		}
+
 		noise_set_static_identity_private_key(&wg->static_identity, in_device.private_key);
 		modified_static_identity = true;
 	}
@@ -161,9 +184,9 @@ int config_set_device(struct wireguard_device *wg, void __user *user_device)
 	for (i = 0, offset = 0, user_peer = user_device + sizeof(struct wgdevice); i < in_device.num_peers; ++i, user_peer += offset) {
 		ret = set_peer(wg, user_peer, &offset);
 		if (ret)
-			break;
+			goto out;
 	}
-
+	ret = 0;
 out:
 	mutex_unlock(&wg->device_update_lock);
 	memzero_explicit(&in_device.private_key, NOISE_PUBLIC_KEY_LEN);
@@ -186,20 +209,6 @@ static inline int use_data(struct data_remaining *data, size_t size)
 	return 0;
 }
 
-static int calculate_ipmasks_size(void *ctx, struct wireguard_peer *peer, union nf_inet_addr ip, u8 cidr, int family)
-{
-	size_t *count = ctx;
-	*count += sizeof(struct wgipmask);
-	return 0;
-}
-
-static size_t calculate_peers_size(struct wireguard_device *wg)
-{
-	size_t len = peer_total_count(wg) * sizeof(struct wgpeer);
-	routing_table_walk_ips(&wg->peer_routing_table, &len, calculate_ipmasks_size);
-	return len;
-}
-
 static int populate_ipmask(void *ctx, union nf_inet_addr ip, u8 cidr, int family)
 {
 	int ret;
@@ -266,9 +275,9 @@ static int populate_peer(struct wireguard_peer *peer, void *ctx)
 	return ret;
 }
 
-int config_get_device(struct wireguard_device *wg, void __user *udevice)
+int config_get_device(struct wireguard_device *wg, void __user *user_device)
 {
-	int ret = 0;
+	int ret;
 	struct net_device *dev = netdev_pub(wg);
 	struct data_remaining peer_data = { NULL };
 	struct wgdevice out_device;
@@ -281,16 +290,21 @@ int config_get_device(struct wireguard_device *wg, void __user *udevice)
 
 	mutex_lock(&wg->device_update_lock);
 
-	if (!udevice) {
-		ret = calculate_peers_size(wg);
+	if (!user_device) {
+		ret = peer_total_count(wg) * sizeof(struct wgpeer)
+		    + routing_table_count_nodes(&wg->peer_routing_table) * sizeof(struct wgipmask);
 		goto out;
 	}
 
-	if (copy_from_user(&in_device, udevice, sizeof(in_device))) {
-		ret = -EFAULT;
+	ret = -EFAULT;
+	if (copy_from_user(&in_device, user_device, sizeof(in_device)))
 		goto out;
-	}
 
+	ret = -EPROTO;
+	if (in_device.version_magic != WG_API_VERSION_MAGIC)
+		goto out;
+
+	out_device.version_magic = WG_API_VERSION_MAGIC;
 	out_device.port = wg->incoming_port;
 	out_device.fwmark = wg->fwmark;
 	strncpy(out_device.interface, dev->name, IFNAMSIZ - 1);
@@ -300,19 +314,23 @@ int config_get_device(struct wireguard_device *wg, void __user *udevice)
 	if (wg->static_identity.has_identity) {
 		memcpy(out_device.private_key, wg->static_identity.static_private, WG_KEY_LEN);
 		memcpy(out_device.public_key, wg->static_identity.static_public, WG_KEY_LEN);
-		memcpy(out_device.preshared_key, wg->static_identity.preshared_key, WG_KEY_LEN);
 	}
+	if (wg->static_identity.has_psk)
+		memcpy(out_device.preshared_key, wg->static_identity.preshared_key, WG_KEY_LEN);
 	up_read(&wg->static_identity.lock);
 
 	peer_data.out_len = in_device.peers_size;
-	peer_data.data = udevice + sizeof(struct wgdevice);
+	peer_data.data = user_device + sizeof(struct wgdevice);
 	ret = peer_for_each_unlocked(wg, populate_peer, &peer_data);
 	if (ret)
 		goto out;
 	out_device.num_peers = peer_data.count;
 
-	if (copy_to_user(udevice, &out_device, sizeof(out_device)))
-		ret = -EFAULT;
+	ret = -EFAULT;
+	if (copy_to_user(user_device, &out_device, sizeof(out_device)))
+		goto out;
+
+	ret = 0;
 
 out:
 	mutex_unlock(&wg->device_update_lock);
